# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

import itertools
from collections.abc import Iterable

from gitfourchette.qt import *
from gitfourchette.search.searchprovider import SearchProvider


class ItemViewSearchProvider(SearchProvider):
    _buddy: QAbstractItemView

    def __init__(self, parent):
        super().__init__(parent)
        assert isinstance(parent, QAbstractItemView)
        self._buddy = parent
        self.dataRole = Qt.ItemDataRole.DisplayRole
        self._startHint = (-1, True)

    @property
    def buddyModel(self):
        # To filter out hidden rows, don't use _buddy.clModel directly
        return self._buddy.model()

    def _currentRow(self):
        return self._buddy.currentIndex().row() if self._buddy.selectedIndexes() else -1

    def _walk(self, forward: bool, jump: bool, skipCurrent: bool, startHint: int = -1):
        """
        Do not override!
        """

        assert self._status != self.TermStatus.Loading

        # -------------------
        # Get row generator

        numRows = self.buddyModel.rowCount()
        currentRow = self._currentRow()

        # Find start bound of search range
        if startHint >= 0:
            start = startHint
        elif currentRow >= 0:
            start = currentRow
        elif forward:
            start = 0
        else:
            start = numRows - 1

        if skipCurrent and start == currentRow:
            start += 1 if forward else -1
            start %= numRows

        if forward:
            range1 = range(start, numRows)
            range2 = range(0, start)
        else:
            range1 = range(start, -1, -1)
            range2 = range(numRows - 1, start, -1)

        rows = itertools.chain(range1, range2)

        # -------------------
        # Walk the model

        try:
            index = self._walkModelImpl(rows)
        except KeyError:
            self._enshrineBadTerm()
            return -1

        # -------------------
        # We've got a matching index

        assert index.isValid()
        self.setStatus(SearchProvider.TermStatus.Good)

        if jump:
            self._jumpToIndex(index)

        return index.row()

    def _jumpToIndex(self, index: QModelIndex):
        # Select the index via QItemSelectionModel, not _buddy.setCurrentIndex,
        # to avoid extending the selection if Ctrl is held down (e.g. pasting a
        # search term with Ctrl+V).
        sm = self._buddy.selectionModel()
        sm.setCurrentIndex(index, QItemSelectionModel.SelectionFlag.SelectCurrent)

    def _walkModelImpl(self, rows: Iterable[int]) -> QModelIndex:
        """
        Iterate on the buddy's model until a matching row is found.
        Raise KeyError if not found.
        Your provider must either set a valid dataRole, or override this method.
        """

        assert self.dataRole != Qt.ItemDataRole.DisplayRole, "set up a proper data role for search"
        assert self._term, "don't search for an empty term"
        assert self._term == self._term.lower(), "search term should have been sanitized"
        assert self._status != self.TermStatus.Loading

        model = self.buddyModel
        for i in rows:
            index = model.index(i, 0)
            data = model.data(index, self.dataRole)
            if data and self._term in data.lower():
                return index

        raise KeyError()

    # -------------------------------------------------------------------------
    # SearchProvider implementation

    def _termChanged(self):
        # Repaint buddy to refresh any highlighting
        self._buddy.viewport().update()

    def prime(self, forwardHint: bool):
        i = self._walk(forwardHint, False, False, -1)
        self._startHint = (i, forwardHint)

    def jump(self, forward: bool):
        assert self.isGoodAndNonEmpty()
        rowHint, rowHintForward = self._startHint
        if not (rowHint >= 0 and rowHintForward == forward):
            rowHint = -1
        rowHint = self._walk(forward, True, True, rowHint)
        self._startHint = (rowHint, forward)

    def invalidate(self):
        super().invalidate()
        self._startHint = (-1, True)

    def isCurrentMatch(self) -> bool:
        row = self._currentRow()
        if row < 0:
            return False
        try:
            self._walkModelImpl([row])
            return True
        except KeyError:
            return False



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

import logging
import os
from collections.abc import Callable
from pathlib import Path

from gitfourchette.localization import *
from gitfourchette.qt import *
from gitfourchette.tasks.repotask import RepoTask, AbortTask
from gitfourchette.toolbox import *

logger = logging.getLogger(__name__)


class NewRepo(RepoTask):
    def flow(self, openRepo: Callable[[str], None]):
        fileDialog = PersistentFileDialog.saveFile(self.parentWidget(), "NewRepo", _("New repository"))
        fileDialog.setFileMode(QFileDialog.FileMode.Directory)
        fileDialog.setLabelText(QFileDialog.DialogLabel.Accept, _("&Create repo here"))
        fileDialog.setWindowModality(Qt.WindowModality.WindowModal)
        pathStr = yield from self.flowFileDialog(fileDialog)
        path = Path(pathStr)

        # macOS's native file picker may return a directory that doesn't exist yet
        # (it expects us to create it ourselves). "git rev-parse" won't detect the
        # parent repo if the directory doesn't exist.
        parentDetectionPath = path if path.exists() else path.parent

        # Discover parent repo. If found, ask user if they want to open the
        # parent repo or create a subrepo inside it.
        revParse = yield from self.flowCallGit(
            "rev-parse",
            "--show-toplevel",
            workdir=str(parentDetectionPath),
            autoFail=False)

        if revParse.exitCode() == 0:
            parentPathStr = revParse.stdoutScrollback().rstrip()
            parentPath = Path(parentPathStr)
            parentWorkdir = parentPath.parent if parentPath.name == ".git" else parentPath
            wantOpen = yield from self._confirmOpenParentRepo(path, parentPath, parentWorkdir)
            if wantOpen:
                openRepo(str(parentWorkdir))
                raise AbortTask()

        # Confirm whether to create .git in existing source tree (non-empty directory)
        if path.exists() and any(path.iterdir()):
            message = _("Are you sure you want to initialize a Git repository in {0}? "
                        "This directory isn’t empty.", bquo(str(path)))
            yield from self.flowConfirm(
                title=_("Directory isn’t empty"),
                text=message,
                verb=_("&Create repo here"),
                icon="warning")

        yield from self.flowCallGit("init", "--", str(path))
        openRepo(str(path))

    def _confirmOpenParentRepo(
            self,
            path: Path,
            parentPath: Path,
            parentWorkdir: Path
    ) -> RepoTask.Flow[bool]:
        """
        If the user is attempting to initialize a repo nested inside a parent
        repo, ask them if they'd rather open the parent repo.
        Return True to open the parent repo, False to create a nested repo.
        """

        path = path.resolve()
        parentPath = parentPath.resolve()
        parentWorkdir = parentWorkdir.resolve()

        if parentPath == path or parentWorkdir == path:
            message = paragraphs(
                _("A repository already exists here:"),
                escape(compactPath(parentWorkdir)))
            yield from self.flowConfirm(
                title=_("Repository already exists"),
                text=message,
                verb=_("&Open existing repo"),
                buttonIcon="SP_DirOpenIcon")
            wantOpen = True

        else:
            displayParentName = parentWorkdir.name
            displayName = path.name
            displayPath = compactPath(path)
            commonLength = len(os.path.commonprefix([displayPath, compactPath(parentWorkdir)]))
            i1 = commonLength - len(displayParentName)
            i2 = commonLength
            dp1 = escape(displayPath[: i1])
            dp2 = escape(displayPath[i1: i2])
            dp3 = escape(displayPath[i2:])
            muted = mutedTextColorHex(self.parentWidget())
            prettyPath = (f"<div style='white-space: pre;'>"
                          f"<span style='color: {muted};'>{dp1}</span>"
                          f"<b>{dp2}</b>"
                          f"<span style='color: {muted};'>{dp3}</span></div>")

            message = paragraphs(
                _("An existing repository, {0}, was found in a parent folder "
                  "of this location:", bquoe(displayParentName)),
                prettyPath,
                _("Are you sure you want to create {0} within the existing "
                  "repo?", hquoe(displayName)))

            createButton = QPushButton(_("&Create {0}", lquoe(displayName)))

            result = yield from self.flowConfirm(
                title=_("Repository found in parent folder"),
                text=message,
                verb=_("&Open {0}", lquoe(displayParentName)),
                buttonIcon="SP_DirOpenIcon",
                actionButton=createButton)

            # OK button is Open; Create is the other one. (Cancel would have aborted early.)
            # Also checking for Accepted so that unit tests can do qmb.accept().
            wantOpen = result in [QMessageBox.StandardButton.Ok, QDialog.DialogCode.Accepted]

        return wantOpen



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

import dataclasses
import enum
import logging
import os
from collections.abc import Iterator
from contextlib import suppress
from typing import TypedDict, ClassVar

from gitfourchette import colors
from gitfourchette.exttools.toolcommands import ToolCommands
from gitfourchette.exttools.toolpresets import ToolPresets
from gitfourchette.localization import *
from gitfourchette.prefsfile import PrefsFile
from gitfourchette.qt import *
from gitfourchette.syntax import PygmentsPresets, ColorScheme
from gitfourchette.toolbox.benchmark import BENCHMARK_LOGGING_LEVEL
from gitfourchette.toolbox.gitutils import AuthorDisplayStyle
from gitfourchette.toolbox.pathutils import PathDisplayStyle
from gitfourchette.toolbox.textutils import englishTitleCase

logger = logging.getLogger(__name__)


SHORT_DATE_PRESETS = {
    "ISO": "yyyy-MM-dd HH:mm",
    "Universal 1": "dd MMM yyyy HH:mm",
    "Universal 2": "ddd dd MMM yyyy HH:mm",
    "European 1": "dd/MM/yy HH:mm",
    "European 2": "dd.MM.yy HH:mm",
    "American": "M/d/yy h:mm ap",
}

SHORT_DATE_DEFAULT_PRESET = next(iter(SHORT_DATE_PRESETS.values()))


class RefSort(enum.IntEnum):
    TimeDesc = 0
    TimeAsc = 1
    AlphaAsc = 2
    AlphaDesc = 3
    UseGlobalPref = -1


class GraphRowHeight(enum.IntEnum):
    Cramped = 80
    Tight = 100
    Relaxed = 130
    Roomy = 150
    Spacious = 175


class GraphRefBoxWidth(enum.IntEnum):
    IconsOnly = 0
    Standard = 120
    Wide = 1000


class QtApiNames(enum.StrEnum):
    Automatic = ""
    PyQt6 = "pyqt6"
    PySide6 = "pyside6"
    PyQt5 = "pyqt5"


class LoggingLevel(enum.IntEnum):
    Benchmark = BENCHMARK_LOGGING_LEVEL
    Debug = logging.DEBUG
    Info = logging.INFO
    Warning = logging.WARNING


class FileListClick(enum.StrEnum):
    Nothing = ""
    Stage = "stage"
    Blame = "blame"
    Edit = "edit"
    DiffTool = "difftool"
    Folder = "folder"


class TabBarClick(enum.StrEnum):
    Nothing = ""
    Close = "close"
    Folder = "folder"
    Terminal = "terminal"


class WhitespaceMode(enum.StrEnum):
    """How `git diff` treats line endings and whitespace when showing patches."""

    Strict = ""
    IgnoreAll = "ignore-all-space"
    IgnoreChange = "ignore-space-change"
    IgnoreCrAtEol = "ignore-cr-at-eol"


@dataclasses.dataclass
class Prefs(PrefsFile):
    _filename = "prefs.json"

    _category_general           : int                   = 0
    language                    : str                   = ""
    qtStyle                     : str                   = ""
    pathDisplayStyle            : PathDisplayStyle      = PathDisplayStyle.FullPaths
    refSort                     : RefSort               = RefSort.TimeDesc
    showToolBar                 : bool                  = True
    showStatusBar               : bool                  = True
    showMenuBar                 : bool                  = True

    _category_diff              : int                   = 0
    font                        : str                   = ""
    fontSize                    : int                   = 0
    syntaxHighlighting          : str                   = PygmentsPresets.Automatic
    colorblind                  : bool                  = False
    contextLines                : int                   = 3
    tabSpaces                   : int                   = 4
    largeFileThresholdKB        : int                   = 500
    wordWrap                    : bool                  = False
    showStrayCRs                : bool                  = True
    showWhitespace              : bool                  = False
    whitespaceMode              : WhitespaceMode        = WhitespaceMode.Strict

    _category_imageDiff         : int                   = 0
    imageFileThresholdKB        : int                   = 5000
    renderSvg                   : bool                  = False

    _category_graph             : int                   = 0
    chronologicalOrder          : bool                  = True
    graphRowHeight              : GraphRowHeight        = GraphRowHeight.Relaxed
    refBoxMaxWidth              : GraphRefBoxWidth      = GraphRefBoxWidth.Standard
    authorDisplayStyle          : AuthorDisplayStyle    = AuthorDisplayStyle.FullName
    shortTimeFormat             : str                   = SHORT_DATE_DEFAULT_PRESET
    maxCommits                  : int                   = 10000
    authorDiffAsterisk          : bool                  = True
    verifyGpgOnTheFly           : bool                  = False
    alternatingRowColors        : bool                  = False

    _category_git               : int                   = 0
    gitPath                     : str                   = ToolPresets.defaultGit()
    ownSshAgent                 : bool                  = False
    ownAskpass                  : bool                  = True
    lfsAware                    : bool                  = True

    _category_external          : int                   = 0
    externalEditor              : str                   = ""
    terminal                    : str                   = ToolPresets.DefaultTerminalCommand
    _spacer0                    : int                   = 0
    externalDiff                : str                   = ToolPresets.DefaultDiffCommand
    externalMerge               : str                   = ToolPresets.DefaultMergeCommand

    _category_userCommands      : int                   = 0
    commands                    : str                   = ""
    confirmCommands             : bool                  = True

    _category_tabs              : int                   = 0
    tabCloseButton              : bool                  = True
    expandingTabs               : bool                  = True
    autoHideTabs                : bool                  = False

    _category_mouseShortcuts    : int                   = 0
    _label_tabBarClicks         : int                   = 0
    doubleClickTabBar           : TabBarClick           = TabBarClick.Folder
    middleClickTabBar           : TabBarClick           = TabBarClick.Close
    _label_fileListClicks       : int                   = 0
    doubleClickFileList         : FileListClick         = FileListClick.Nothing
    middleClickFileList         : FileListClick         = FileListClick.Stage
    _label_diffViewClicks       : int                   = 0
    middleClickStageLines       : bool                  = True

    _category_trash             : int                   = 0
    maxTrashFiles               : int                   = 250
    maxTrashFileKB              : int                   = 1000

    _category_advanced          : int                   = 0
    maxRecentRepos              : int                   = 20
    shortHashChars              : int                   = 7
    autoRefresh                 : bool                  = True
    autoFetchMinutes            : int                   = 5
    flattenLanes                : bool                  = True
    animations                  : bool                  = True
    condensedFonts              : bool                  = True
    pygmentsPlugins             : bool                  = False
    verbosity                   : LoggingLevel          = LoggingLevel.Debug if APP_TESTMODE else LoggingLevel.Warning
    forceQtApi                  : QtApiNames            = QtApiNames.Automatic

    _category_hidden            : int                   = 0
    # Hide autoFetch from PrefsDialog because autoFetchMinutes's control includes a checkbox
    autoFetch                   : bool                  = False
    rememberPassphrases         : bool                  = True
    smoothScroll                : bool                  = True
    toolBarButtonStyle          : Qt.ToolButtonStyle    = Qt.ToolButtonStyle.ToolButtonTextBesideIcon
    toolBarIconSize             : int                   = 16
    defaultCloneLocation        : str                   = ""
    dontShowAgain               : list[str]             = dataclasses.field(default_factory=list)
    resetDontShowAgain          : bool                  = False
    donatePrompt                : int                   = 0
    refSortClearTimestamp       : int                   = 0

    @property
    def listViewScrollMode(self) -> QAbstractItemView.ScrollMode:
        if self.smoothScroll:
            return QAbstractItemView.ScrollMode.ScrollPerPixel
        else:
            return QAbstractItemView.ScrollMode.ScrollPerItem

    def resolveDefaultCloneLocation(self):
        if self.defaultCloneLocation:
            return self.defaultCloneLocation

        path = QStandardPaths.writableLocation(QStandardPaths.StandardLocation.DownloadLocation)
        if path:
            return os.path.normpath(path)
        return os.path.expanduser("~")

    def monoFont(self):
        monoFont = QFontDatabase.systemFont(QFontDatabase.SystemFont.FixedFont)
        if self.font:
            monoFont.fromString(self.font)
        if self.fontSize > 0:
            monoFont.setPointSize(self.fontSize)
        return monoFont

    def isSyntaxHighlightingEnabled(self):
        return self.syntaxHighlighting != PygmentsPresets.Off

    def syntaxHighlightingScheme(self):
        return ColorScheme.resolve(self.syntaxHighlighting)

    def addDelColors(self):
        if self.colorblind:
            return colors.teal, colors.orange
        else:
            return colors.olive, colors.red

    def addDelColorsStyleTag(self):
        green, red = self.addDelColors()
        return ("<style>"
                f"del {{ color: {red.name()}; text-decoration: none; }} "
                f"add {{ color: {green.name()}; }}"
                "</style>")

    def isGitSandboxed(self):
        return self.gitPath.startswith(ToolCommands.FlatpakSandboxedCommandPrefix)


class PrefEffects:
    RebuildMenu: ClassVar = {
        "language",
        "commands",
        "confirmCommands",
        "externalEditor"
    }
    "Pref keys that trigger a rebuild of the main menu."

    ReloadDiff: ClassVar = {
        "showStrayCRs",
        "colorblind",
        "largeFileThresholdKB",
        "imageFileThresholdKB",
        "contextLines",
        "whitespaceMode",
        "maxCommits",
        "renderSvg",
        "lfsAware",
        "syntaxHighlighting",
    }
    "Pref keys that trigger a reload of the current diff."

    ReloadRepo: ClassVar = {
        "chronologicalOrder",
        "maxCommits",
        "refSort",
    }
    "Pref keys that fully take effect after a repo reload."

    RestartApp: ClassVar = {
        "language",
        "forceQtApi",
        "pygmentsPlugins",
    }
    "Pref keys that fully take effect after an app restart."


@dataclasses.dataclass
class History(PrefsFile):
    _filename = "history.json"

    class JsonRepo(TypedDict, total=False):
        seq: int
        length: int
        nickname: str
        superproject: str

    repos: dict[str, JsonRepo] = dataclasses.field(default_factory=dict)
    cloneHistory: list[str] = dataclasses.field(default_factory=list)
    fileDialogPaths: dict[str, str] = dataclasses.field(default_factory=dict)
    startups: int = 0

    _maxSeq = -1

    def addRepo(self, path: str):
        path = os.path.normpath(path)
        repo = self.getRepo(path)
        repo['seq'] = self.drawSequenceNumber()
        return repo

    def getRepo(self, path: str) -> JsonRepo:
        path = os.path.normpath(path)
        try:
            repo = self.repos[path]
        except KeyError:
            repo = History.JsonRepo()
            self.repos[path] = repo
        return repo

    def getRepoNickname(self, path: str, strict: bool = False) -> str:
        repo = self.getRepo(path)
        path = os.path.normpath(path)
        return repo.get("nickname", "" if strict else os.path.basename(path))

    def setRepoNickname(self, path: str, nickname: str):
        repo = self.getRepo(path)
        nickname = nickname.strip()
        if nickname:
            repo['nickname'] = nickname
        else:
            repo.pop('nickname', None)

    def getRepoNumCommits(self, path: str) -> int:
        repo = self.getRepo(path)
        return repo.get('length', 0)

    def setRepoNumCommits(self, path: str, commitCount: int):
        repo = self.getRepo(path)
        if commitCount > 0:
            repo['length'] = commitCount
        else:
            repo.pop('length', None)

    def getRepoSuperproject(self, path: str) -> str:
        repo = self.getRepo(path)
        return repo.get('superproject', "")

    def setRepoSuperproject(self, path: str, superprojectPath: str):
        repo = self.getRepo(path)
        if superprojectPath:
            repo['superproject'] = superprojectPath
        else:
            repo.pop('superproject', None)

    def getRepoTabName(self, path: str) -> str:
        name = self.getRepoNickname(path)

        seen = {path}
        while path:
            path = self.getRepoSuperproject(path)
            if path:
                if path in seen:
                    logger.warning(f"Circular superproject in {self._filename}! {path}")
                    return name
                seen.add(path)
                superprojectName = self.getRepoNickname(path)
                name = f"{superprojectName}: {name}"

        return name

    def removeRepo(self, path: str):
        path = os.path.normpath(path)
        self.repos.pop(path, None)
        self.invalidateSequenceNumber()

    def clearRepoHistory(self):
        self.repos.clear()
        self.invalidateSequenceNumber()

    def getRecentRepoPaths(self, n: int, newestFirst=True) -> Iterator[str]:
        sortedPaths = (path for path, _ in
                       sorted(self.repos.items(), key=lambda i: i[1].get('seq', -1), reverse=newestFirst))

        return (path for path, _ in zip(sortedPaths, range(n), strict=False))

    def write(self, force=False):
        self.trim()
        super().write(force)

    def trim(self):
        n = prefs.maxRecentRepos

        if len(self.repos) > n:
            # Recreate self.repos with only the n most recent paths
            topPaths = self.getRecentRepoPaths(n)
            self.repos = {path: self.repos[path] for path in topPaths}

        if len(self.cloneHistory) > n:
            self.cloneHistory = self.cloneHistory[-n:]

    def addCloneUrl(self, url):
        with suppress(ValueError):
            self.cloneHistory.remove(url)
        # Insert most recent cloned URL first
        self.cloneHistory.insert(0, url)

    def clearCloneHistory(self):
        self.cloneHistory.clear()

    def drawSequenceNumber(self, increment=1):
        if self._maxSeq < 0 and self.repos:
            self._maxSeq = max(r.get('seq', -1) for r in self.repos.values())
        self._maxSeq += increment
        return self._maxSeq

    def invalidateSequenceNumber(self):
        self._maxSeq = -1


@dataclasses.dataclass
class Session(PrefsFile):
    _filename = "session.json"

    tabs                        : list[str]             = dataclasses.field(default_factory=list)
    activeTabIndex              : int                   = -1
    windowGeometry              : bytes                 = b""
    splitterSizes               : dict[str, list[int]]  = dataclasses.field(default_factory=dict)


# Initialize default prefs and history.
# The app should load the user's prefs with prefs.load() and history.load().
prefs = Prefs()
history = History()


def qtIsNativeMacosStyle():  # pragma: no cover
    if not MACOS:
        return False
    return prefs.qtStyle.lower() == "macos"


def getExternalEditorName():
    genericName = englishTitleCase(_("External editor"))
    return ToolPresets.getCommandName(prefs.externalEditor, genericName, ToolPresets.Editors)


def getDiffToolName():
    genericName = englishTitleCase(_("Diff tool"))
    return ToolPresets.getCommandName(prefs.externalDiff, genericName, ToolPresets.DiffTools)


def getMergeToolName():
    genericName = englishTitleCase(_("Merge tool"))
    return ToolPresets.getCommandName(prefs.externalMerge, genericName, ToolPresets.MergeTools)



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

import itertools
import math
import os
from collections.abc import Callable
from pathlib import Path

from gitfourchette.qt import *

QModelIndex_default = QModelIndex()
QPoint_zero = QPoint()

ShortcutKeys = QKeySequence | QKeySequence.StandardKey | Qt.Key | str
MultiShortcut = list[QKeySequence]

SupportedImageSuffixes = {'.' + fmt.data().decode('ascii', errors='replace')
                          for fmt in QImageReader.supportedImageFormats()}


def openFolder(path: str):
    QDesktopServices.openUrl(QUrl.fromLocalFile(path))


def showInFolder(path: str):  # pragma: no cover (platform-specific)
    """
    Show a file or folder with explorer/finder.
    Source for Windows & macOS: https://stackoverflow.com/a/46019091/3388962
    """
    path = os.path.abspath(path)
    isdir = os.path.isdir(path)

    if APP_TESTMODE:
        pass

    elif FREEDESKTOP and HAS_QTDBUS:
        # https://www.freedesktop.org/wiki/Specifications/file-manager-interface
        iface = QDBusInterface("org.freedesktop.FileManager1", "/org/freedesktop/FileManager1")
        if iface.isValid():
            if PYQT5 or PYQT6:
                # PyQt5/6 needs the array of strings to be spelled out explicitly.
                if PYQT5:
                    stringType = QMetaType.Type.QString
                else:
                    stringType = QMetaType.Type.QString.value
                dbusArgs = QDBusArgument()
                dbusArgs.beginArray(stringType)  # type: ignore[call-overload]  # inaccurate stubs?
                dbusArgs.add(path)
                dbusArgs.endArray()
            else:
                # Thankfully, PySide6 is more pythonic here.
                dbusArgs = [path]  # type: ignore[assignment]  # consider only PyQt6 for type checking
            iface.call("ShowItems", dbusArgs, "")
            iface.deleteLater()
            return

    elif WINDOWS:
        if not isdir:  # If it's a file, select it within the folder.
            args = ['/select,', path]
        else:
            args = [path]  # If it's a folder, open it.
        if QProcess.startDetached('explorer', args):
            return

    elif MACOS and not isdir:
        args = [
            '-e', 'tell application "Finder"',
            '-e', 'activate',
            '-e', F'select POSIX file "{path}"',
            '-e', 'end tell',
            '-e', 'return'
        ]
        if not QProcess.execute('/usr/bin/osascript', args):
            return

    # Fallback.
    dirPath = path if os.path.isdir(path) else os.path.dirname(path)
    openFolder(dirPath)


def onAppThread():
    appInstance = QApplication.instance()
    return bool(appInstance and appInstance.thread() is QThread.currentThread())


def isImageFormatSupported(filename: str):
    """
    Guesses whether an image is in a supported format from its filename.
    This is for when QImageReader.imageFormat(path) doesn't cut it (e.g. if the file doesn't exist on disk).
    """
    ext = Path(filename).suffix.lower()
    return ext in SupportedImageSuffixes


def setFontFeature(font: QFont, fourCC: str, value: int = 1):
    try:
        font.setFeature(QFont.Tag(fourCC), value)
    except AttributeError:  # pragma: no cover
        # Mitigation for pre-Qt 6.7 bindings
        pass
    return font


def adjustedWidgetFontSize(widget: QWidget, relativeSize: int = 100):
    return round(widget.font().pointSize() * relativeSize / 100.0)


def tweakWidgetFont(
        widget: QWidget,
        relativeSize: int = 100,
        bold: bool = False,
        tabularNumbers: bool = False
):
    font: QFont = widget.font()
    if relativeSize != 100:
        font.setPointSize(round(font.pointSize() * relativeSize / 100.0))
    if bold:
        font.setBold(bold)
    if tabularNumbers:
        setFontFeature(font, "tnum")
    widget.setFont(font)
    return font


def formatWidgetText(widget: QAbstractButton | QLabel, *args, **kwargs):
    text = widget.text()
    text = text.format(*args, **kwargs)
    widget.setText(text)
    return text


def formatWidgetTooltip(widget: QWidget, forceWrap=False, *args, **kwargs):
    text = widget.toolTip()
    text = text.format(*args, **kwargs)
    if forceWrap:
        text = "<html>" + text
    widget.setToolTip(text)
    return text


def enforceComboBoxMaxVisibleItems(comboBox: QComboBox, maxItems=0):
    """
    Some styles, in particular Fusion and macOS, ignore QComboBox.maxVisibleItems
    (see https://doc.qt.io/qt-6/qcombobox.html#maxVisibleItems-prop).

    This is an issue because if the popup spans the entire height the screen,
    Fusion doesn't account for taskbars, permanent menubars, etc.
    So, the first/last items may remain off-screen.

    This function enforces maxVisibleItems on a QComboBox if the style is bad.
    """

    if maxItems > 0:
        comboBox.setMaxVisibleItems(maxItems)

    isBadStyle = comboBox.style().styleHint(QStyle.StyleHint.SH_ComboBox_Popup, QStyleOptionComboBox())
    if not isBadStyle:
        return

    # QStyleSheetStyle::styleHint() (qstylesheetstyle.cpp)
    comboBox.setStyleSheet(comboBox.styleSheet() + "\nQComboBox { combobox-popup: 0; }")

    listView = comboBox.view()
    assert isinstance(listView, QListView)
    listView.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAsNeeded)


def reevaluateStyleSheet(widget: QWidget):
    qss = widget.styleSheet()
    if not qss:
        qss = "/* dummy */"
    widget.setStyleSheet(qss)


def toggleQssProperty(widget: QWidget, name: str, value: bool):
    if value == bool(widget.property(name)):
        return

    widget.setProperty(name, "true" if value else None)

    # Reset stylesheet to percolate property change
    reevaluateStyleSheet(widget)


def isDarkTheme(palette: QPalette | None = None):
    if palette is None:
        palette = QApplication.palette()
    themeBG = palette.color(QPalette.ColorRole.Base)  # standard theme background color
    themeFG = palette.color(QPalette.ColorRole.Text)  # standard theme foreground color
    return themeBG.lightness() < themeFG.lightness()


def mutedTextColorHex(w: QWidget, alpha=.5) -> str:
    mutedColor = QApplication.palette().windowText().color()
    mutedColor.setAlphaF(alpha)
    return mutedColor.name(QColor.NameFormat.HexArgb)


def mutedToolTipColorHex() -> str:
    mutedColor = QApplication.palette().toolTipText().color()
    mutedColor.setAlphaF(.6)
    return mutedColor.name(QColor.NameFormat.HexArgb)


def appendShortcutToToolTipText(tip: str, shortcut: ShortcutKeys, singleLine=True):
    if not isinstance(shortcut, QKeySequence):
        shortcut = QKeySequence(shortcut)

    hint = shortcut.toString(QKeySequence.SequenceFormat.NativeText)
    hint = f"<span style='color: {mutedToolTipColorHex()}'> &nbsp;{hint}</span>"
    prefix = ""
    if singleLine:
        prefix = "<p style='white-space: pre'>"
    return f"{prefix}{tip} {hint}"


def appendShortcutToToolTip(widget: QWidget, shortcut: ShortcutKeys, singleLine=True):
    tip = widget.toolTip()
    tip = appendShortcutToToolTipText(tip, shortcut, singleLine)
    widget.setToolTip(tip)
    return tip


def itemViewVisibleRowRange(view: QAbstractItemView):
    assert isinstance(view, QListView)
    model = view.model()  # use the view's top-level model to only search filtered rows

    rect = view.viewport().contentsRect()
    top = view.indexAt(rect.topLeft())
    if not top.isValid():
        return range(-1)

    bottom = view.indexAt(rect.bottomLeft())
    if not bottom.isValid():
        bottom = model.index(model.rowCount() - 1, 0)

    return range(top.row(), bottom.row() + 1)


class DisableWidgetUpdatesContext:
    def __init__(self, widget: QWidget):
        self.widget = widget

    def __enter__(self):
        self.widget.setUpdatesEnabled(False)

    def __exit__(self, excType, excValue, excTraceback):
        self.widget.setUpdatesEnabled(True)

    @staticmethod
    def methodDecorator(func: Callable) -> Callable:
        def wrapper(*args, **kwargs) -> Callable:
            widget: QWidget = args[0]
            with DisableWidgetUpdatesContext(widget):
                return func(*args, **kwargs)
        return wrapper


class MakeNonNativeDialog(QObject):  # pragma: no cover (macOS-specific)
    """
    Enables the AA_DontUseNativeDialogs attribute, and disables it when the dialog is shown.
    Meant to be used to disable the iOS-like styling of dialog boxes on modern macOS.
    """
    def __init__(self, parent: QDialog):
        super().__init__(parent)
        nonNativeAlready = QCoreApplication.testAttribute(Qt.ApplicationAttribute.AA_DontUseNativeDialogs)
        if nonNativeAlready:
            self.deleteLater()
            return
        QApplication.setAttribute(Qt.ApplicationAttribute.AA_DontUseNativeDialogs, True)
        parent.installEventFilter(self)

    def eventFilter(self, watched, event: QEvent):
        if event.type() == QEvent.Type.Show:
            QCoreApplication.setAttribute(Qt.ApplicationAttribute.AA_DontUseNativeDialogs, False)
            watched.removeEventFilter(self)
            self.deleteLater()
        return False


class QScrollBackupContext:
    def __init__(self, *items: QAbstractScrollArea | QScrollBar):
        self.scrollBars: list[QScrollBar] = []
        self.values: list[int] = []

        for o in items:
            if isinstance(o, QAbstractScrollArea):
                self.scrollBars.append(o.horizontalScrollBar())
                self.scrollBars.append(o.verticalScrollBar())
            else:
                assert isinstance(o, QScrollBar)
                self.scrollBars.append(o)

    def __enter__(self):
        self.values = [o.value() for o in self.scrollBars]

    def __exit__(self, exc_type, exc_val, exc_tb):
        for o, v in zip(self.scrollBars, self.values, strict=True):
            o.setValue(v)


class CallbackAccumulator(QTimer):
    def __init__(self, parent: QObject, callback: Callable, delay: int = 0):
        super().__init__(parent)
        self.setObjectName("CallbackAccumulator")
        self.setSingleShot(True)
        self.setInterval(delay)
        self.timeout.connect(callback)

    @staticmethod
    def deferredMethod(delay: int = 0):
        assert not callable(delay), "did you forget parentheses in '@deferredMethod()'?"

        def decorator(callback: Callable):
            attr = f"__callbackaccumulator_{id(callback)}"

            def wrapper(obj):
                try:
                    defer = getattr(obj, attr)
                except AttributeError:
                    defer = CallbackAccumulator(obj, lambda: callback(obj), delay)
                    setattr(obj, attr, defer)
                defer.start()

            return wrapper

        return decorator


def makeInternalLink(urlAuthority: str, urlPath: str = "", urlFragment: str = "", **urlQueryItems) -> str:
    url = QUrl()
    url.setScheme(APP_URL_SCHEME)
    url.setAuthority(urlAuthority)

    if urlPath:
        if not urlPath.startswith("/"):
            urlPath = "/" + urlPath
        url.setPath(urlPath)

    if urlFragment:
        url.setFragment(urlFragment)

    query = QUrlQuery()
    for k, v in urlQueryItems.items():
        query.addQueryItem(k, v)
    if query:
        url.setQuery(query)

    return url.toString()


def makeMultiShortcut(*args: str | QKeySequence.StandardKey | Qt.Key | QKeySequence) -> MultiShortcut:
    shortcuts: list[QKeySequence] = []

    for alt in args:
        if isinstance(alt, QKeySequence.StandardKey):
            shortcuts.extend(QKeySequence.keyBindings(alt))
        elif isinstance(alt, (str, Qt.Key)):
            shortcuts.append(QKeySequence(alt))
        else:
            assert isinstance(alt, QKeySequence)
            shortcuts.append(alt)

    # Ensure no duplicates (stable order since Python 3.7+)
    shortcuts = list(dict.fromkeys(shortcuts))

    return shortcuts


def makeWidgetShortcut(
        parent: QWidget,
        callback: Callable | SignalInstance,
        *keys: ShortcutKeys,
        context: Qt.ShortcutContext = Qt.ShortcutContext.WidgetShortcut,
) -> QShortcut:
    assert keys, "no shortcut keys given"
    shortcut = QShortcut(parent)
    if QT5:  # pragma: no cover - Only one key per shortcut in Qt 5
        shortcut.setKey(keys[0])
    else:
        shortcut.setKeys(keys)
    shortcut.setContext(context)
    shortcut.activated.connect(callback)
    return shortcut


def lerp(v1, v2, c=.5, cmin=0.0, cmax=1.0):
    p = (c-cmin) / (cmax-cmin)
    p = max(p, 0)
    p = min(p, 1)
    v = v2*p + v1*(1-p)
    return v


def mixColors(c1: QColor, c2: QColor, ratio=.5, rmin=0.0, rmax=1.0):
    return QColor.fromRgbF(
        lerp(c1.redF(),   c2.redF(),   ratio, rmin, rmax),
        lerp(c1.greenF(), c2.greenF(), ratio, rmin, rmax),
        lerp(c1.blueF(),  c2.blueF(),  ratio, rmin, rmax),
        lerp(c1.alphaF(), c2.alphaF(), ratio, rmin, rmax))


def relativeLuminance(color: QColor) -> float:
    """Relative luminance per WCAG 2.2"""
    # https://www.w3.org/TR/WCAG21/relative-luminance.html
    def srgbToLinear(x: float):
        return x/12.92 if x <= .04045 else math.pow((x+.055)/ 1.055, 2.4)
    r = .2126 * srgbToLinear(color.redF())
    g = .7152 * srgbToLinear(color.greenF())
    b = .0722 * srgbToLinear(color.blueF())
    return r + g + b


def findParentWidget(o: QObject) -> QWidget:
    p = o.parent()
    while p:
        if isinstance(p, QWidget):
            return p
        p = p.parent()
    raise ValueError(f"No parent widget found for {o!r}")


def setTabOrder(*args: QWidget):
    """
    Qt 6.6 introduces QWidget::setTabOrder(std::initializer_list<QWidget *> widgets)
    but neither PySide6 nor PyQt6 expose it yet.
    """
    for widget1, widget2 in itertools.pairwise(args):
        QWidget.setTabOrder(widget1, widget2)


def packDialog(dialog: QDialog, widthHint=550, lockHeight=False):
    dialog.layout().activate()
    dialog.adjustSize()

    if widthHint:
        dialog.resize(max(widthHint, dialog.width()), dialog.height())

    if lockHeight:
        dialog.setFixedHeight(dialog.height())


class DocumentLinks:
    """
    Bundle of ad-hoc links bound to callback functions.
    """

    UrlAuthority = "adhoc"

    def __init__(self):
        self.callbacks = {}

    def new(self, callback: Callable[[], None]) -> str:
        key = QUuid.createUuid().toString(QUuid.StringFormat.WithoutBraces)
        self.callbacks[key] = callback
        return makeInternalLink(self.UrlAuthority, urlFragment=key)

    def processLink(self, url: QUrl | str) -> bool:
        # Convert to QUrl so we can extract elements from it
        if isinstance(url, str):
            url = QUrl(url)

        # Let something else process this link if it's not for us
        if url.scheme() != APP_URL_SCHEME or url.authority() != self.UrlAuthority:
            return False

        key = url.fragment()
        callback = self.callbacks[key]
        callback()
        return True



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from gitfourchette.filelists.filelist import FileList
from gitfourchette.gitdriver import GitDelta
from gitfourchette.globalshortcuts import GlobalShortcuts
from gitfourchette.localization import *
from gitfourchette.nav import NavContext
from gitfourchette.porcelain import *
from gitfourchette.qt import *
from gitfourchette.repomodel import RepoModel
from gitfourchette.tasks import *
from gitfourchette.toolbox import *


class DirtyFiles(FileList):
    def __init__(self, repoModel: RepoModel, parent: QWidget):
        super().__init__(repoModel, parent, NavContext.UNSTAGED)

        self.setSelectionMode(QAbstractItemView.SelectionMode.ExtendedSelection)

        makeWidgetShortcut(self, self.stage, *GlobalShortcuts.stageHotkeys)
        makeWidgetShortcut(self, self.discard, *GlobalShortcuts.discardHotkeys)

    def contextMenuActions(self, deltas: list[GitDelta]) -> list[ActionDef]:
        actions = []

        n = len(deltas)

        statusSet = {delta.status for delta in deltas}
        modeSet = {delta.new.mode for delta in deltas}  # mode in worktree

        anyConflicts = "U" in statusSet  # DeltaStatus.CONFLICTED in statusSet
        anySubmodules = FileMode.COMMIT in modeSet
        onlyConflicts = anyConflicts and len(statusSet) == 1
        onlySubmodules = anySubmodules and len(modeSet) == 1
        onlyUntracked = statusSet == {"?"}  # {DeltaStatus.UNTRACKED}

        if not anyConflicts and not anySubmodules:
            contextMenuActionStage = ActionDef(
                _n("&Stage File", "&Stage {n} Files", n),
                self.stage,
                icon="git-stage",
                shortcuts=GlobalShortcuts.stageHotkeys[0])

            contextMenuActionIgnore = ActionDef(
                _n("&Ignore Untracked File…", "&Ignore Untracked Files…", n),
                self.ignoreSelection,
                enabled=n == 1 and onlyUntracked)

            if onlyUntracked:
                discardText = _n("&Delete File…", "&Delete {n} Files…", n)
            else:
                discardText = _n("&Discard Changes in File…", "&Discard Changes in {n} Files…", n)
            contextMenuActionDiscard = ActionDef(
                discardText,
                self.discard,
                icon="git-discard",
                shortcuts=GlobalShortcuts.discardHotkeys[0])

            actions += [
                contextMenuActionStage,
                contextMenuActionDiscard,
                self.contextMenuActionStash(),
                self.contextMenuActionRevertMode(deltas, self.discardModeChanges),
                contextMenuActionIgnore,
                ActionDef.SEPARATOR,
                self.contextMenuActionBlame(deltas),
                *self.contextMenuActionsDiff(deltas),
            ]

        elif onlyConflicts:
            actions += [
                ActionDef(
                    _n("Merge Conflict", "{n} Merge Conflicts", n),
                    kind=ActionDef.Kind.Section,
                ),
                ActionDef(_("Keep OUR version") + "\u2026", lambda: self.hardSolve(True)),
                ActionDef(_("Accept THEIR version") + "\u2026", lambda: self.hardSolve(False)),
            ]

        elif onlySubmodules:
            actions += [
                ActionDef(
                    _n("Submodule", "{n} Submodules", n),
                    kind=ActionDef.Kind.Section,
                ),
                ActionDef(
                    _n("Stage Submodule", "Stage {n} Submodules", n),
                    self.stage,
                ),
                ActionDef(
                    _n("Discard Changes in Submodule", "Discard Changes in {n} Submodules", n),
                    self.discard,
                ),
                ActionDef(
                    _n("Open Submodule in New Tab", "Open {n} Submodules in New Tabs", n),
                    self.openSubmoduleTabs,
                ),
            ]

        else:
            # Conflicted + non-conflicted files selected
            # or Submodules + non-submodules selected
            sorry = _("Can’t stage this selection in bulk.") + "\n" + _("Please review the files individually.")
            actions += [
                ActionDef(sorry, enabled=False),
            ]

        if actions:
            actions.append(ActionDef.SEPARATOR)

        if not onlySubmodules:
            actions += [
                *self.contextMenuActionsEdit(deltas),
                ActionDef.SEPARATOR,
            ]

        actions += super().contextMenuActions(deltas)
        return actions

    def wantStageOrUnstage(self):
        self.stage()

    def stage(self):
        deltas = list(self.selectedDeltas())
        StageFiles.invoke(self, deltas)

    def discard(self):
        deltas = list(self.selectedDeltas())
        DiscardFiles.invoke(self, deltas)

    def discardModeChanges(self):
        deltas = list(self.selectedDeltas())
        DiscardModeChanges.invoke(self, deltas)

    def hardSolve(self, keepOurs: bool):
        conflicts = [delta.conflict for delta in self.selectedDeltas()]
        HardSolveConflicts.invoke(self, conflicts, keepOurs)

    def ignoreSelection(self):
        selected = list(self.selectedPaths())
        assert len(selected) == 1, "QAction bound to ignoreSelection should be disabled"
        NewIgnorePattern.invoke(self, selected[0])
